/* This code is (C) AllegroGL contributors, and double licensed under
 * the GPL and zlib licenses. See gpl.txt or zlib.txt for details.
 */
/** \file videovtb.c
  * \brief video bitmaps (ie. texture rendering) vtable
  *  Some of these don't work correctly or will be very slow.
  *  The current vtable is in early stages, and will be replaced by a better
  *  (faster) version 'soon'. Don't count on regular Allegro functions
  *  working on the screen or on video bitmaps.
  */

#include <string.h>
#include <limits.h>

#include <allegro.h>

#ifdef ALLEGRO_WINDOWS
#include <winalleg.h>
#endif

#include "alleggl.h"
#include "allglint.h"
#include "glvtable.h"
#include <allegro/internal/aintern.h>
#ifdef ALLEGRO_MACOSX
#include <OpenGL/glu.h>
#else
#include <GL/glu.h>
#endif


static GFX_VTABLE allegro_gl_video_vtable;

/* Counter for video bitmaps. screen = 1 */
static int video_bitmap_count = 2;

static int __allegro_gl_video_bitmap_bpp = -1;


void allegro_gl_destroy_video_bitmap(BITMAP *bmp);



static int allegro_gl_make_video_bitmap_helper1(int w, int h, int x, int y,
                                   GLint target, AGL_VIDEO_BITMAP **pvid) {

	int internal_format;
	int bpp;
   
	if (__allegro_gl_video_bitmap_bpp == -1) {
		bpp = bitmap_color_depth(screen);
	}
	else {
		bpp = __allegro_gl_video_bitmap_bpp;
	}

	(*pvid) = malloc(sizeof(AGL_VIDEO_BITMAP));
			
	if (!(*pvid))
		return -1;
				
	memset(*pvid, 0, sizeof(AGL_VIDEO_BITMAP));
		
	/* Create associated bitmap */
	(*pvid)->memory_copy = create_bitmap_ex(bpp, w, h);
	if (!(*pvid)->memory_copy)
		return -1;
	
	(*pvid)->format = __allegro_gl_get_bitmap_color_format((*pvid)->memory_copy, AGL_TEXTURE_HAS_ALPHA);
	(*pvid)->type = __allegro_gl_get_bitmap_type((*pvid)->memory_copy, 0);
	internal_format = __allegro_gl_get_texture_format_ex((*pvid)->memory_copy, AGL_TEXTURE_HAS_ALPHA);

	(*pvid)->target = target;

	/* Fill in some values in the bitmap to make it act as a subbitmap
	 */
	(*pvid)->width  = w;
	(*pvid)->height = h;
	(*pvid)->x_ofs = x;
	(*pvid)->y_ofs = y;

	/* Make a texture out of it */
	glGenTextures(1, &((*pvid)->tex));
	if (!((*pvid)->tex))
		return -1;

	glEnable((*pvid)->target);
	glBindTexture((*pvid)->target, ((*pvid)->tex));

	glTexImage2D((*pvid)->target, 0, internal_format, w, h,
	             0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);

	/* By default, use the Allegro filtering mode - ie: Nearest */
	glTexParameteri((*pvid)->target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri((*pvid)->target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	/* Clamp to edge */
	{	GLenum clamp = GL_CLAMP_TO_EDGE;
		if (!allegro_gl_extensions_GL.SGIS_texture_edge_clamp) {
			clamp = GL_CLAMP;
		}
		glTexParameteri((*pvid)->target, GL_TEXTURE_WRAP_S, clamp);
		glTexParameteri((*pvid)->target, GL_TEXTURE_WRAP_T, clamp);
	}
	glDisable((*pvid)->target);

	return 0;
}



static int allegro_gl_make_video_bitmap_helper0(int w, int h, int x, int y,
                                             	AGL_VIDEO_BITMAP **pvid) {
		
	int is_power_of_2 = (!(w & (w - 1)) && !(h & (h - 1)));
	int texture_rect_available = allegro_gl_extensions_GL.ARB_texture_rectangle
#ifdef ALLEGRO_MACOSX
							 || allegro_gl_extensions_GL.EXT_texture_rectangle
#endif
							 || allegro_gl_extensions_GL.NV_texture_rectangle;
	GLint max_rect_texture_size = 0;

	if (texture_rect_available) {
		glGetIntegerv(GL_MAX_RECTANGLE_TEXTURE_SIZE_ARB, &max_rect_texture_size);
	}

	if (w <= allegro_gl_info.max_texture_size &&
		h <= allegro_gl_info.max_texture_size) {
		if (allegro_gl_extensions_GL.ARB_texture_non_power_of_two ||
			is_power_of_2) {
			if (allegro_gl_make_video_bitmap_helper1(w, h, x, y,
											GL_TEXTURE_2D, pvid)) {
				return -1;
			}
		}
		else if (texture_rect_available &&
				 w <= max_rect_texture_size &&
				 h <= max_rect_texture_size) {
			if (allegro_gl_make_video_bitmap_helper1(w, h, x, y,
											GL_TEXTURE_RECTANGLE_ARB, pvid)) {
				return -1;
			}
		}
		else {
			/* NPO2 textures are not suppored by the driver in any way.
			 * Split the bitmap into smaller POT bitmaps. */
			const unsigned int BITS = sizeof(int) * CHAR_BIT;
			unsigned int i, j;
			unsigned int w1, h1;
			unsigned int x1, y1;
			unsigned int p1, p2;

			/* Find the POTs using bits. */
			y1 = 0;
			for (i = 0; i < BITS; i++) {
				p1 = 1 << i;
				if (h & p1)
					h1 = p1;
				else
					continue;

				x1 = 0;
				for (j = 0; j < BITS; j++) {
					p2 = 1 << j;
					if (w & p2)
						w1 = p2;
					else
						continue;

					if (allegro_gl_make_video_bitmap_helper0(w1, h1, x + x1,
														y + y1, pvid)) {
						return -1;
					}
					do {
						pvid = &((*pvid)->next);
					} while (*pvid);

					x1 += w1;
				}

				y1 += h1;
			}
		}
	}
	else {
		/* Texture is too big to fit. Split it in 4 and try again. */
		int w1, w2, h1, h2;

		w2 = w / 2;
		w1 = w - w2;

		h2 = h / 2;
		h1 = h - h2;

		/* Even a 1x1 texture didn't work? Bail*/
		if (!w2 && !h2) {
			return -1;
		}

		/* Top-left corner */
		if (allegro_gl_make_video_bitmap_helper0(w1, h1, x, y, pvid)) {
			return -1;
		}
		do {
			pvid = &((*pvid)->next);
		} while (*pvid);

		/* Top-right corner */
		if (w2) {
			if (allegro_gl_make_video_bitmap_helper0(w2, h1, x + w1, y, pvid)) {
				return -1;
			}
			do {
				pvid = &((*pvid)->next);
			} while (*pvid);
		}
		/* Bottom-left corner */
		if (h2) {
			if (allegro_gl_make_video_bitmap_helper0(w1, h2, x, y + h1, pvid)) {
				return -1;
			}
			do {
				pvid = &((*pvid)->next);
			} while (*pvid);
		}
		/* Bottom-right corner */
		if (w2 && h2) {
			if (allegro_gl_make_video_bitmap_helper0(w2, h2, x + w1, y + h1, pvid)) {
				return -1;
			}
			do {
				pvid = &((*pvid)->next);
			} while (*pvid);
		}
	}

	return 0;
}



/* Will make a (potentially piece-wise) texture out of the specified bitmap
 * Source -must- be a memory bitmap or memory subbitmap (created by Allegro
 * only).
 *
 */
static BITMAP *allegro_gl_make_video_bitmap(BITMAP *bmp) {
	
	/* Grab a pointer to the bitmap data to patch */
	void *ptr = &bmp->extra;
	AGL_VIDEO_BITMAP **pvid = (AGL_VIDEO_BITMAP**)ptr;
	
	/* Convert bitmap to texture */
	if (allegro_gl_make_video_bitmap_helper0(bmp->w, bmp->h, 0, 0, pvid)) {
		goto agl_error;
	}

	return bmp;
	
agl_error:
	allegro_gl_destroy_video_bitmap(bmp);
	return NULL;
}



/* void allegro_gl_destroy_video_bitmap(BITMAP *bmp) */
/** destroy_video_bitmap() overload.
  * Will destroy the video bitmap. You mustn't use the BITMAP pointer after
  * calling this function.
  */
void allegro_gl_destroy_video_bitmap(BITMAP *bmp) {

	AGL_VIDEO_BITMAP *vid = bmp ? bmp->extra : NULL, *next;
	
	if (!bmp)
		return;
	
	while (vid) {
		if (vid->memory_copy)
			destroy_bitmap(vid->memory_copy);
				
		if (vid->tex)
			glDeleteTextures(1, &vid->tex);
								
		next = vid->next;
		free(vid);
		vid = next;
	}

	free(bmp->vtable);
	free(bmp);
	
	return;
}



/* BITMAP *allegro_gl_create_video_bitmap(int w, int h) */
/** create_video_bitmap() overload.
  * This function will create a video bitmap using texture memory.
  * Video bitmaps do not currently share space with the frame buffer (screen).
  * Video bitmaps are lost when switching screen modes.
  */
BITMAP *allegro_gl_create_video_bitmap(int w, int h) {
	GFX_VTABLE *vtable;
	BITMAP *bitmap;

	bitmap = malloc(sizeof(BITMAP) + sizeof(char *));
	
	if (!bitmap)
		return NULL;

	bitmap->dat = NULL;
	bitmap->w = bitmap->cr = w;
	bitmap->h = bitmap->cb = h;
	bitmap->clip = TRUE;
	bitmap->cl = bitmap->ct = 0;
	bitmap->write_bank = bitmap->read_bank = NULL;
	/* We should keep tracks of allocated bitmaps for the ref counter */
	bitmap->id = BMP_ID_VIDEO | video_bitmap_count;
	bitmap->extra = NULL;
	bitmap->x_ofs = 0;
	bitmap->y_ofs = 0;
	bitmap->seg = _default_ds();
	bitmap->line[0] = NULL;

	if (!allegro_gl_make_video_bitmap(bitmap)) {
		return NULL;
	}
	video_bitmap_count++;
	
	/* XXX <rohannessian> We ought to leave the Allegro values intact
	 * Avoids bad interaction with correct Allegro programs.
	 */
	vtable = malloc(sizeof(struct GFX_VTABLE));
	*vtable = allegro_gl_video_vtable;
	if (__allegro_gl_video_bitmap_bpp == -1) {
		vtable->color_depth = bitmap_color_depth(screen);
	}
	else {
		vtable->color_depth = __allegro_gl_video_bitmap_bpp;
	}
	switch (vtable->color_depth) {
		case 8:
			vtable->mask_color = MASK_COLOR_8;
		break;
		case 15:
			vtable->mask_color = MASK_COLOR_15;
		break;
		case 16:
			vtable->mask_color = MASK_COLOR_16;
		break;
		case 24:
			vtable->mask_color = MASK_COLOR_24;
		break;
		case 32:
			vtable->mask_color = MASK_COLOR_32;
		break;
	}
	bitmap->vtable = vtable;

	return bitmap;
}



/* allegro_gl_set_video_bitmap_color_depth(int bpp) */
/** \ingroup bitmap
 *  Sets the color depth you'd like AllegroGL to use for video bitmaps.
 *  This function lets you create video bitmaps of any color depth using
 *  allegro's create_video_bitmap(). You can create a RGBA8 bitmap and take
 *  advantage of hardware accelerated blending, no matter what the current color
 *  depth is.
 *  This call will affect all subsequent calls to create_video_bitmap().
 *
 *  Default value for the video bitmap color depth if the color depth of the
 *  screen bitmap. To revert to the default color depth, pass -1.
 *
 *  \param bpp Color depth in btis per pixel.
 *  \return The previous value of the video bitmap color depth.
 */
GLint allegro_gl_set_video_bitmap_color_depth(int bpp) {
	GLint old_val = __allegro_gl_video_bitmap_bpp;
	__allegro_gl_video_bitmap_bpp = bpp;
	return old_val;
}


/* static void allegro_gl_video_acquire(struct BITMAP *bmp) */
/** \ingroup glvtable
  * acquire_bitmap(bmp) overload.
  * This doesn't do anything, since OpenGL textures
  * doesn't need locking. You don't need to call this function
  * in your program.
  */
static void allegro_gl_video_acquire(struct BITMAP *bmp) {}



/* static void allegro_gl_video_release(struct BITMAP *bmp) */
/** \ingroup glvtable
  * release_bitmap(bmp) overload.
  * This doesn't do anything, since OpenGL textures
  * doesn't need locking. You don't need to call this function
  * in your program.
  */
static void allegro_gl_video_release(struct BITMAP *bmp) {}



static int allegro_gl_video_getpixel(struct BITMAP *bmp, int x, int y)
{
	int pix = -1;
	AGL_VIDEO_BITMAP *vid;
	AGL_LOG(2, "glvtable.c:allegro_gl_screen_getpixel\n");	
	
	if (is_sub_bitmap(bmp)) {
		x += bmp->x_ofs;
		y += bmp->y_ofs;
	}
	if (x < bmp->cl || x >= bmp->cr || y < bmp->ct || y >= bmp->cb) {
		return -1;
	}

	vid = bmp->extra;
	
	while (vid) {
		if (vid->x_ofs <= x && vid->y_ofs <= y
		 && vid->x_ofs + vid->memory_copy->w > x
		 && vid->y_ofs + vid->memory_copy->h > y) {
			
			pix = getpixel(vid->memory_copy, x - vid->x_ofs, y - vid->y_ofs);
			break;
		}
		vid = vid->next;
	}
	
	if (pix != -1) {
		return pix;
	}
	
	return -1;
}



static void allegro_gl_video_putpixel(struct BITMAP *bmp, int x, int y,
                                      int color) {
	AGL_VIDEO_BITMAP *vid;

	if (is_sub_bitmap(bmp)) {
		x += bmp->x_ofs;
		y += bmp->y_ofs;
	}
	if (x < bmp->cl || x >= bmp->cr || y < bmp->ct || y >= bmp->cb) {
		return;
	}

	vid = bmp->extra;

	while (vid) {
		if (vid->x_ofs <= x && vid->y_ofs <= y
		 && vid->x_ofs + vid->memory_copy->w > x
		 && vid->y_ofs + vid->memory_copy->h > y) {
			
			putpixel(vid->memory_copy, x - vid->x_ofs, y - vid->y_ofs, color);
			
			glEnable(vid->target);
			glBindTexture(vid->target, vid->tex);
			glTexSubImage2D(vid->target, 0, x - vid->x_ofs,
				y - vid->y_ofs, 1, 1, vid->format, vid->type, &color);
			glBindTexture(vid->target, 0);
			glDisable(vid->target);
				
			break;
		}
		vid = vid->next;
	}
	
	return;
}



static void allegro_gl_video_vline(BITMAP *bmp, int x, int y1, int y2,
                                   int color) {

	AGL_VIDEO_BITMAP *vid;
	GLint saved_row_length;
	GLint saved_alignment;
	int bpp;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_vline\n");
	vid = bmp->extra;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	glGetIntegerv(GL_UNPACK_ALIGNMENT, &saved_alignment);
	
	if (is_sub_bitmap(bmp)) {
		x  += bmp->x_ofs;
		y1 += bmp->y_ofs;
		y2 += bmp->y_ofs;
	}
	if (x < bmp->cl || x >= bmp->cr) {
		return;
	}
	
	if (y1 > y2) {
		int temp = y1;
		y1 = y2;
		y2 = temp;
	}

	if (y1 < bmp->ct) {
		y1 = bmp->ct;
	}
	if (y2 >= bmp->cb) {
		y2 = bmp->cb - 1;
	}

	bpp = BYTES_PER_PIXEL(bitmap_color_depth(vid->memory_copy));
	
	while (vid) {
		BITMAP *vbmp = vid->memory_copy;
		
		int _y1, _y2, _x;
		if (vid->x_ofs > x || vid->y_ofs > y2
		 || vid->x_ofs + vbmp->w <= x
		 || vid->y_ofs + vbmp->h <= y1) {
			
			vid = vid->next;
			continue;
		}
		
		_y1 = MAX(y1, vid->y_ofs) - vid->y_ofs;
		_y2 = MIN(y2, vid->y_ofs + vbmp->h - 1)	- vid->y_ofs;
		_x = x - vid->x_ofs;

		vline(vbmp, _x, _y1, _y2, color);

		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		              vbmp->h > 1
		             ? (vbmp->line[1] - vbmp->line[0]) / bpp
		             : vbmp->w);

		glEnable(vid->target);
		glBindTexture(vid->target, vid->tex);
		glTexSubImage2D(vid->target, 0, _x, _y1, 1, _y2 - _y1 + 1, vid->format,
			vid->type, vbmp->line[_y1] + _x * bpp);
		glBindTexture(vid->target, 0);
		glDisable(vid->target);

		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glPixelStorei(GL_UNPACK_ALIGNMENT,	saved_alignment);
	
	return;
}



static void allegro_gl_video_hline(BITMAP *bmp, int x1, int y, int x2,
                                   int color) {

	AGL_VIDEO_BITMAP *vid;
	GLint saved_row_length;
	GLint saved_alignment;
	int bpp;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_hline\n");
	vid = bmp->extra;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	glGetIntegerv(GL_UNPACK_ALIGNMENT, &saved_alignment);

	if (is_sub_bitmap(bmp)) {
		x1 += bmp->x_ofs;
		x2 += bmp->x_ofs;
		y  += bmp->y_ofs;
	}

	if (y < bmp->ct || y >= bmp->cb) {
		return;
	}
	
	if (x1 > x2) {
		int temp = x1;
		x1 = x2;
		x2 = temp;
	}

	if (x1 < bmp->cl) {
		x1 = bmp->cl;
	}
	if (x2 >= bmp->cr) {
		x2 = bmp->cr - 1;
	}

	bpp = BYTES_PER_PIXEL(bitmap_color_depth(vid->memory_copy));

	while (vid) {
		BITMAP *vbmp = vid->memory_copy;
		
		int _x1, _x2, _y;
		if (vid->y_ofs > y || vid->x_ofs > x2
		 || vid->x_ofs + vbmp->w <= x1
		 || vid->y_ofs + vbmp->h <= y) {
			
			vid = vid->next;
			continue;
		}
		
		_x1 = MAX(x1, vid->x_ofs) - vid->x_ofs;
		_x2 = MIN(x2, vid->x_ofs + vbmp->w - 1)	- vid->x_ofs;
		_y = y - vid->y_ofs;

		hline(vbmp, _x1, _y, _x2, color);
	
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		              vbmp->h > 1
		             ? (vbmp->line[1] - vbmp->line[0]) / bpp
		             : vbmp->w);
		
		glEnable(vid->target);
		glBindTexture(vid->target, vid->tex);
		glTexSubImage2D(vid->target, 0, _x1, _y, _x2 - _x1 + 1, 1, vid->format,
			vid->type, vbmp->line[_y] + _x1 * bpp);
		glBindTexture(vid->target, 0);
		glDisable(vid->target);

		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glPixelStorei(GL_UNPACK_ALIGNMENT,	saved_alignment);
	
	return;
}



static void allegro_gl_video_line(struct BITMAP *bmp, int x1, int y1, int x2,
                                  int y2, int color) {
	
	/* Note: very very slow */						
	do_line(bmp, x1, y1, x2, y2, color, allegro_gl_video_putpixel);
	
	return;
}
	


static void allegro_gl_video_rectfill(struct BITMAP *bmp, int x1, int y1,
                                      int x2, int y2, int color) {

	AGL_VIDEO_BITMAP *vid;
	GLint saved_row_length;
	GLint saved_alignment;
	int bpp;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_rectfill\n");
	vid = bmp->extra;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	glGetIntegerv(GL_UNPACK_ALIGNMENT, &saved_alignment);
	
	if (is_sub_bitmap(bmp)) {
		x1 += bmp->x_ofs;
		x2 += bmp->x_ofs;
		y1 += bmp->y_ofs;
		y2 += bmp->y_ofs;
	}
	
	if (y1 > y2) {
		int temp = y1;
		y1 = y2;
		y2 = temp;
	}

	if (x1 > x2) {
		int temp = x1;
		x1 = x2;
		x2 = temp;
	}

	if (x1 < bmp->cl) {
		x1 = bmp->cl;
	}
	if (x2 >= bmp->cr) {
		x2 = bmp->cr - 1;
	}
	if (y1 < bmp->ct) {
		y1 = bmp->ct;
	}
	if (y1 >= bmp->cb) {
		y1 = bmp->cb - 1;
	}

	bpp = BYTES_PER_PIXEL(bitmap_color_depth(vid->memory_copy));

	while (vid) {
		BITMAP *vbmp = vid->memory_copy;
		
		int _y1, _y2, _x1, _x2;
		if (vid->x_ofs > x2 || vid->y_ofs > y2
		 || vid->x_ofs + vbmp->w <= x1
		 || vid->y_ofs + vbmp->h <= y1) {
			
			vid = vid->next;
			continue;
		}
		
		_y1 = MAX(y1, vid->y_ofs) - vid->y_ofs;
		_y2 = MIN(y2, vid->y_ofs + vbmp->h - 1)	- vid->y_ofs;
		_x1 = MAX(x1, vid->x_ofs) - vid->x_ofs;
		_x2 = MIN(x2, vid->x_ofs + vbmp->w - 1)	- vid->x_ofs;

		rectfill(vbmp, _x1, _y1, _x2, _y2, color);
		
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		              vbmp->h > 1
		             ? (vbmp->line[1] - vbmp->line[0]) / bpp
		             : vbmp->w);

		glEnable(vid->target);
		glBindTexture(vid->target, vid->tex);
		glTexSubImage2D(vid->target, 0,
		    _x1, _y1, _x2 - _x1 + 1, _y2 - _y1 + 1, vid->format,
			vid->type, vbmp->line[_y1] + _x1 * bpp);
		glBindTexture(vid->target, 0);
		glDisable(vid->target);

		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glPixelStorei(GL_UNPACK_ALIGNMENT,	saved_alignment);
	
	return;
}


static void allegro_gl_video_triangle(struct BITMAP *bmp, int x1, int y1,
                                      int x2, int y2, int x3, int y3, int color)
{	
	AGL_VIDEO_BITMAP *vid;
	GLint saved_row_length;
	GLint saved_alignment;
	int min_y, max_y, min_x, max_x;
	int bpp;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_triangle\n");
	vid = bmp->extra;
	
	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	glGetIntegerv(GL_UNPACK_ALIGNMENT, &saved_alignment);

	if (is_sub_bitmap(bmp)) {
		x1 += bmp->x_ofs;
		x2 += bmp->x_ofs;
		x3 += bmp->x_ofs;
		y1 += bmp->y_ofs;
		y2 += bmp->y_ofs;
		y3 += bmp->y_ofs;
	}

	min_y = MIN(y1, MIN(y2, y3));
	min_x = MIN(x1, MIN(x2, x3));
	max_y = MAX(y1, MAX(y2, y3));
	max_x = MAX(x1, MAX(x2, x3));

	bpp = BYTES_PER_PIXEL(bitmap_color_depth(vid->memory_copy));

	while (vid) {
		BITMAP *vbmp = vid->memory_copy;
		
		int _y1, _y2, _x1, _x2, _x3, _y3;
		if (vid->x_ofs > max_x || vid->y_ofs > max_y
		 || vid->x_ofs + vbmp->w <= min_x
		 || vid->y_ofs + vbmp->h <= min_y) {
			
			vid = vid->next;
			continue;
		}
		
		_y1 = y1 - vid->y_ofs;
		_y2 = y2 - vid->y_ofs;
		_y3 = y3 - vid->y_ofs;
		_x1 = x1 - vid->x_ofs;
		_x2 = x2 - vid->x_ofs;
		_x3 = x3 - vid->x_ofs;


		set_clip_rect(vbmp, bmp->cl - vid->x_ofs, bmp->ct - vid->y_ofs,
		                    bmp->cr - vid->x_ofs - 1, bmp->cb - vid->y_ofs - 1);

		triangle(vbmp, _x1, _y1, _x2, _y2, _x3, _y3, color);

		set_clip_rect(vbmp, 0, 0, vbmp->w - 1, vbmp->h - 1);
		
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		              vbmp->h > 1
		             ? (vbmp->line[1] - vbmp->line[0]) / bpp
		             : vbmp->w);
		
		/* Not quite the minimal rectangle occupied by the triangle, but
		 * pretty close
		 */
		_y1 = MAX(0, min_y - vid->y_ofs);
		_y2 = MIN(vbmp->h - 1, max_y - vid->y_ofs);
		_x1 = MAX(0, min_x - vid->x_ofs);
		_x2 = MIN(vbmp->w - 1, max_x - vid->x_ofs);
			
		glEnable(vid->target);
		glBindTexture(vid->target, vid->tex);
		glTexSubImage2D(vid->target, 0,
		    _x1, _y1, _x2 - _x1 + 1, _y2 - _y1 + 1, vid->format,
			vid->type, vbmp->line[_y1] + _x1 * bpp);
		glBindTexture(vid->target, 0);
		glDisable(vid->target);

		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glPixelStorei(GL_UNPACK_ALIGNMENT,	saved_alignment);
}



void allegro_gl_video_blit_from_memory(struct BITMAP *source,
                struct BITMAP *dest, int source_x, int source_y,
				int dest_x, int dest_y, int width, int height) {

	AGL_VIDEO_BITMAP *vid;
	BITMAP *dest_parent = dest;
	GLint saved_row_length;
	GLint saved_alignment;
	int bpp;
	GLenum format, type;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_blit_from_memory\n");

	if (is_sub_bitmap (dest)) {
	   dest_x += dest->x_ofs;
	   dest_y += dest->y_ofs;
	   while (dest_parent->id & BMP_ID_SUB)
	      dest_parent = (BITMAP *)dest_parent->extra;
	}

	if (dest_x < dest->cl) {
		dest_x = dest->cl;
	}
	if (dest_y < dest->ct) {
		dest_y = dest->ct;
	}
	if (dest_x + width >= dest->cr) {
		width = dest->cr - dest_x;
	}
	if (dest_y + height >= dest->cb) {
		height = dest->cb - dest_y;
	}
	if (width < 1 || height < 1) {
		return;
	}
	
	vid = dest_parent->extra;

	bpp = BYTES_PER_PIXEL(bitmap_color_depth(vid->memory_copy));
	format = vid->format;
	type = vid->type;

	glGetIntegerv(GL_UNPACK_ROW_LENGTH, &saved_row_length);
	glGetIntegerv(GL_UNPACK_ALIGNMENT, &saved_alignment);
	
	while (vid) {
		BITMAP *temp = NULL;
		BITMAP *vbmp = vid->memory_copy;

		int _x, _y, _w, _h;
		if (vid->x_ofs >= dest_x + width || vid->y_ofs >= dest_y + height
		 || vid->x_ofs + vbmp->w <= dest_x
		 || vid->y_ofs + vbmp->h <= dest_y) {
			
			vid = vid->next;
			continue;
		}

		_x = MAX (vid->x_ofs, dest_x) - vid->x_ofs;
		_w = MIN (vid->x_ofs + vbmp->w, dest_x + width)
		   - vid->x_ofs - _x;
		_y = MAX (vid->y_ofs, dest_y) - vid->y_ofs;
		_h = MIN (vid->y_ofs + vbmp->h, dest_y + height)
		   - vid->y_ofs - _y;

		blit(source, vbmp, source_x + vid->x_ofs + _x - dest_x,
		     source_y + vid->y_ofs + _y - dest_y, _x, _y, _w, _h);

		/* If packed pixels (or GL 1.2) isn't supported, then we need to convert
		 * the bitmap into something GL can understand - 24-bpp should do it.
		 */
		if (!allegro_gl_extensions_GL.EXT_packed_pixels
				   						&& bitmap_color_depth(source) < 24) {
			temp = create_bitmap_ex(24, vbmp->w, vbmp->h);
			if (!temp)
				return;

			blit(vbmp, temp, 0, 0, 0, 0, temp->w, temp->h);
			vbmp = temp;
			bpp = BYTES_PER_PIXEL(bitmap_color_depth(vbmp));
			format = __allegro_gl_get_bitmap_color_format(vbmp, 0);
			type = __allegro_gl_get_bitmap_type(vbmp, 0);
		}

		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glPixelStorei(GL_UNPACK_ROW_LENGTH,
		              vbmp->h > 1
		             ? (vbmp->line[1] - vbmp->line[0]) / bpp
		             : vbmp->w);

		glEnable(vid->target);
		glBindTexture(vid->target, vid->tex);
		glTexSubImage2D(vid->target, 0, _x, _y, _w, _h,
			format, type, vbmp->line[_y] + _x * bpp);
		glBindTexture(vid->target, 0);
		glDisable(vid->target);

		if (temp)
			destroy_bitmap(vbmp);
			
		vid = vid->next;
	}
	
	glPixelStorei(GL_UNPACK_ROW_LENGTH,	saved_row_length);
	glPixelStorei(GL_UNPACK_ALIGNMENT,	saved_alignment);

	return;	
}



void allegro_gl_video_blit_to_memory(struct BITMAP *source, struct BITMAP *dest,
                         int source_x, int source_y, int dest_x, int dest_y,
                         int width, int height) {

	AGL_VIDEO_BITMAP *vid;
	BITMAP *source_parent = source;
	
	AGL_LOG(2, "glvtable.c:allegro_gl_video_blit_to_memory\n");
	
	if (is_sub_bitmap(source)) {
	   source_x += source->x_ofs;
	   source_y += source->y_ofs;
	   while (source_parent->id & BMP_ID_SUB)
	      source_parent = (BITMAP *)source_parent->extra;
	}

	vid = source_parent->extra;
	
	while (vid) {
		BITMAP *vbmp = vid->memory_copy;
		int x, y, dx, dy, w, h;

		x = MAX(source_x, vid->x_ofs) - vid->x_ofs;
		y = MAX(source_y, vid->y_ofs) - vid->y_ofs;
		w = MIN(vid->x_ofs + vbmp->w, source_x + width) - vid->x_ofs;
		h = MIN(vid->y_ofs + vbmp->h, source_y + height) - vid->y_ofs;
		dx = MAX(0, vid->x_ofs - source_x) + dest_x;
		dy = MAX(0, vid->y_ofs - source_y) + dest_y;

		blit(vbmp, dest, x, y, dx, dy, w, h);
	
		vid = vid->next;
	}

	return;	
}



static void allegro_gl_video_clear_to_color(BITMAP *bmp, int color) {

	AGL_LOG(2, "glvtable.c:allegro_gl_video_clear_to_color\n");
	
	allegro_gl_video_rectfill(bmp, 0, 0, bmp->w, bmp->h, color);
			
	return;
}



static void dummy_unwrite_bank(void)
{
}



static GFX_VTABLE allegro_gl_video_vtable = {
	0,
	0,
	dummy_unwrite_bank,			//void *unwrite_bank;  /* C function on some machines, asm on i386 */
	NULL,						//AL_METHOD(void, set_clip, (struct BITMAP *bmp));
	allegro_gl_video_acquire,
	allegro_gl_video_release,
	NULL,						//AL_METHOD(struct BITMAP *, create_sub_bitmap, (struct BITMAP *parent, int x, int y, int width, int height));
	allegro_gl_created_sub_bitmap,
	allegro_gl_video_getpixel,
	allegro_gl_video_putpixel,
	allegro_gl_video_vline,
	allegro_gl_video_hline,
	allegro_gl_video_hline,
	allegro_gl_video_line,
	allegro_gl_video_line,
	allegro_gl_video_rectfill,
	allegro_gl_video_triangle,
	NULL,/*allegro_gl_screen_draw_sprite,*/
	NULL,						//AL_METHOD(void, draw_256_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y));
	NULL,/*allegro_gl_screen_draw_sprite_v_flip,*/
	NULL,/*allegro_gl_screen_draw_sprite_h_flip,*/
	NULL,/*allegro_gl_screen_draw_sprite_vh_flip,*/
	NULL,						//AL_METHOD(void, draw_trans_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_trans_rgba_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_lit_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y, int color));
	NULL,						//AL_METHOD(void, draw_rle_sprite, (struct BITMAP *bmp, struct RLE_SPRITE *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_trans_rle_sprite, (struct BITMAP *bmp, struct RLE_SPRITE *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_trans_rgba_rle_sprite, (struct BITMAP *bmp, struct RLE_SPRITE *sprite, int x, int y));
	NULL,						//AL_METHOD(void, draw_lit_rle_sprite, (struct BITMAP *bmp, struct RLE_SPRITE *sprite, int x, int y, int color));
	NULL,						//AL_METHOD(void, draw_character, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y, int color));
	NULL,/*allegro_gl_screen_draw_glyph,*/
	allegro_gl_video_blit_from_memory,
	allegro_gl_video_blit_to_memory,
	NULL,						//AL_METHOD(void, blit_from_system, (struct BITMAP *source, struct BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height));
	NULL,						//AL_METHOD(void, blit_to_system, (struct BITMAP *source, struct BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height));
	allegro_gl_screen_blit_to_self, /* Video bitmaps use same method as screen */
	allegro_gl_screen_blit_to_self, /* ..._forward */
	allegro_gl_screen_blit_to_self, /* ..._backward */
	allegro_gl_memory_blit_between_formats,
	NULL,						//AL_METHOD(void, masked_blit, (struct BITMAP *source, struct BITMAP *dest, int source_x, int source_y, int dest_x, int dest_y, int width, int height));
	allegro_gl_video_clear_to_color,
	NULL,						//AL_METHOD(void, pivot_scaled_sprite_flip, (struct BITMAP *bmp, struct BITMAP *sprite, fixed x, fixed y, fixed cx, fixed cy, fixed angle, fixed scale, int v_flip));
	NULL,                       //AL_METHOD(void, do_stretch_blit, (struct BITMAP *source, struct BITMAP *dest, int source_x, int source_y, int source_width, int source_height, int dest_x, int dest_y, int dest_width, int dest_height, int masked));
	NULL,                       //AL_METHOD(void, draw_gouraud_sprite, (struct BITMAP *bmp, struct BITMAP *sprite, int x, int y, int c1, int c2, int c3, int c4));
	NULL,                       //AL_METHOD(void, draw_sprite_end, (void));
	NULL,                       //AL_METHOD(void, blit_end, (void));
	_soft_polygon,              //AL_METHOD(void, polygon, (struct BITMAP *bmp, int vertices, AL_CONST int *points, int color));
	_soft_rect,                 //AL_METHOD(void, rect, (struct BITMAP *bmp, int x1, int y1, int x2, int y2, int color));
	_soft_circle,               //AL_METHOD(void, circle, (struct BITMAP *bmp, int x, int y, int radius, int color));
	_soft_circlefill,           //AL_METHOD(void, circlefill, (struct BITMAP *bmp, int x, int y, int radius, int color));
	_soft_ellipse,              //AL_METHOD(void, ellipse, (struct BITMAP *bmp, int x, int y, int rx, int ry, int color));
	_soft_ellipsefill,          //AL_METHOD(void, ellipsefill, (struct BITMAP *bmp, int x, int y, int rx, int ry, int color));
	_soft_arc,                  //AL_METHOD(void, arc, (struct BITMAP *bmp, int x, int y, fixed ang1, fixed ang2, int r, int color));
	_soft_spline,               //AL_METHOD(void, spline, (struct BITMAP *bmp, AL_CONST int points[8], int color));
	_soft_floodfill,            //AL_METHOD(void, floodfill, (struct BITMAP *bmp, int x, int y, int color));
	_soft_polygon3d,            //AL_METHOD(void, polygon3d, (struct BITMAP *bmp, int type, struct BITMAP *texture, int vc, V3D *vtx[]));
	_soft_polygon3d_f,          //AL_METHOD(void, polygon3d_f, (struct BITMAP *bmp, int type, struct BITMAP *texture, int vc, V3D_f *vtx[]));
	_soft_triangle3d,           //AL_METHOD(void, triangle3d, (struct BITMAP *bmp, int type, struct BITMAP *texture, V3D *v1, V3D *v2, V3D *v3));
	_soft_triangle3d_f,         //AL_METHOD(void, triangle3d_f, (struct BITMAP *bmp, int type, struct BITMAP *texture, V3D_f *v1, V3D_f *v2, V3D_f *v3));
	_soft_quad3d,               //AL_METHOD(void, quad3d, (struct BITMAP *bmp, int type, struct BITMAP *texture, V3D *v1, V3D *v2, V3D *v3, V3D *v4));
	_soft_quad3d_f,             //AL_METHOD(void, quad3d_f, (struct BITMAP *bmp, int type, struct BITMAP *texture, V3D_f *v1, V3D_f *v2, V3D_f *v3, V3D_f *v4));
};

